"""Defines Database parameters for a catalog."""

from dataclasses import dataclass
from typing import List, Optional, Union


@dataclass(frozen=True)
class JDBCParam:
    """JDBC connection parameters for a database."""

    username: str
    password: str
    jdbc_uri: str
    driver_class: str
    driver_url: str


@dataclass(frozen=True)
class ElasticSearchParam:
    """ElasticSearch connection parameters for a database."""

    hosts: List[str]
    username: Optional[str] = None
    password: Optional[str] = None


DatabaseParams = Union[JDBCParam, ElasticSearchParam]



"""Utility functions for conversion."""

from typing import Any, Dict, List, Union

import dacite
from puppygraph.common.dataclass_utils import dataclass_to_camel_dict
from puppygraph.data.mapping.catalog_config import CatalogConfig, CatalogType
from puppygraph.data.mapping.database_params import (
    DatabaseParams,
    ElasticSearchParam,
    JDBCParam,
)
from puppygraph.data.mapping.datalake_params import (
    DatalakeParams,
    MetastoreParam,
    S3StorageParam,
    UnityMetastoreParam,
)
from puppygraph.data.mapping.graph_element_cache_config import (
    GraphElementCacheConfig,
)
from puppygraph.data.mapping.graph_mapping_config import (
    GraphElementConfig,
    MappedField,
    PuppyGraphMappingConfig,
    TableSource,
)


def convert_mapping_config_to_host_json(
    config: Union[PuppyGraphMappingConfig, Dict],
) -> Dict[str, Any]:
    """Converts the PuppyGraph Mapping config to the host schema JSON.

    Args:
        config: The PuppyGraph Mapping config.

    Returns:
        The host schema JSON.
    """

    def _metastore_param_to_json(metastore_param: MetastoreParam) -> Dict[str, Any]:
        json = {}
        if isinstance(metastore_param, UnityMetastoreParam):
            json = dataclass_to_camel_dict(metastore_param)
            json["type"] = "unity"
            # Rename unityCatalogName to databricksCatalogName
            if "unityCatalogName" in json:
                json["databricksCatalogName"] = json.pop("unityCatalogName")
        return json

    def _storage_param_to_json(storage_param: S3StorageParam) -> Dict[str, Any]:
        json = {}
        if isinstance(storage_param, S3StorageParam):
            json = dataclass_to_camel_dict(storage_param)
            json["type"] = "S3"
        return json

    def _datalake_params_to_json(datalake_params: DatalakeParams) -> Dict[str, Any]:
        json = {}
        if datalake_params.metastore_param is not None:
            json["metastore"] = _metastore_param_to_json(
                datalake_params.metastore_param
            )
        if datalake_params.storage_param is not None:
            json["storage"] = _storage_param_to_json(datalake_params.storage_param)
        return json

    def _database_params_to_json(database_params: DatabaseParams) -> Dict[str, Any]:
        key = ""
        if isinstance(database_params, JDBCParam):
            key = "jdbc"
        elif isinstance(database_params, ElasticSearchParam):
            key = "elasticSearch"
        return {key: dataclass_to_camel_dict(database_params)}

    def _catalog_config_to_json(catalog_config: CatalogConfig) -> Dict[str, Any]:
        """Maps the catalog config to the host catalog JSON.

        Args:
            catalog_config: The catalog configuration to map.

        Returns:
            The host catalog JSON.
        """
        catalog_json = {
            "name": catalog_config.name,
            "type": catalog_config.type.name.lower(),
        }

        if isinstance(catalog_config.params, DatalakeParams):
            catalog_json.update(_datalake_params_to_json(catalog_config.params))
        elif isinstance(catalog_config.params, DatabaseParams):
            catalog_json.update(_database_params_to_json(catalog_config.params))

        return catalog_json

    def _table_source_to_json(table_source: TableSource) -> Dict[str, Any]:
        return {
            "catalog": table_source.catalog_name,
            "schema": table_source.schema_name,
            "table": table_source.table_name,
        }

    def _mapped_id_to_json(mapped_id: List[MappedField]) -> Dict[str, Any]:
        return {
            "fields": [
                {
                    "field": mapped_field.from_field,
                    "alias": mapped_field.name,
                    "type": mapped_field.type,
                }
                for mapped_field in mapped_id
            ]
        }

    def _attribute_to_json(mapped_attribute: MappedField) -> Dict[str, Any]:
        return {
            "alias": mapped_attribute.name,
            "field": mapped_attribute.from_field,
            "type": mapped_attribute.type,
        }

    def _cache_config_to_json(cache_config: GraphElementCacheConfig) -> Dict[str, Any]:
        return dataclass_to_camel_dict(cache_config)

    def _vertex_config_to_json(vertex_config: GraphElementConfig) -> Dict[str, Any]:
        json = {
            "label": vertex_config.label,
            "oneToOne": {
                "tableSource": _table_source_to_json(
                    table_source=vertex_config.table_source
                ),
                "id": _mapped_id_to_json(mapped_id=vertex_config.id),
                "attributes": [
                    _attribute_to_json(
                        mapped_attribute=attribute,
                    )
                    for attribute in vertex_config.attributes
                ],
            },
        }
        if vertex_config.cache_config is not None:
            json["cache"] = _cache_config_to_json(vertex_config.cache_config)
        return json

    def _edge_config_to_json(
        edge_config: GraphElementConfig,
    ) -> Dict[str, Any]:

        json = {
            "label": edge_config.label,
            "fromVertex": edge_config.from_label,
            "toVertex": edge_config.to_label,
            "tableSource": _table_source_to_json(table_source=edge_config.table_source),
            "id": _mapped_id_to_json(mapped_id=edge_config.id),
            "fromId": _mapped_id_to_json(mapped_id=edge_config.from_id),
            "toId": _mapped_id_to_json(mapped_id=edge_config.to_id),
            "attributes": [
                _attribute_to_json(
                    mapped_attribute=attribute,
                )
                for attribute in edge_config.attributes
            ],
        }

        if edge_config.cache_config is not None:
            json["cache"] = _cache_config_to_json(edge_config.cache_config)
        return json

    if isinstance(config, dict):
        config = dacite.from_dict(
            data_class=PuppyGraphMappingConfig,
            data=config,
            config=dacite.Config(
                type_hooks={
                    CatalogType: lambda x: CatalogType[x.upper()],
                }
            ),
        )

    return {
        "catalogs": [
            _catalog_config_to_json(catalog_config)
            for catalog_config in config.catalogs
        ],
        "graph": {
            "vertices": [
                _vertex_config_to_json(vertex_config=vertex_config)
                for vertex_config in config.vertices
            ],
            "edges": [
                _edge_config_to_json(edge_config=edge_config)
                for edge_config in config.edges
            ],
        },
    }



"""PuppyGraph construction config."""

from dataclasses import dataclass, field
from typing import List, Optional

from puppygraph.data.mapping.catalog_config import CatalogConfig
from puppygraph.data.mapping.graph_element_cache_config import (
    GraphElementCacheConfig,
)


@dataclass(frozen=True)
class TableSource:
    """Table source."""

    catalog_name: str
    schema_name: str
    table_name: str


@dataclass(frozen=True)
class MappedField:
    """MappedAttributes from the source table to the graph schema."""

    name: str
    type: str
    from_field: str
    description: Optional[str] = None


@dataclass(frozen=True)
class GraphElementConfig:
    """Graph Element config.

    A graph element is a vertex or an edge.
    """

    table_source: TableSource

    label: str

    # List of attributes, each attribute is a field in the source table
    attributes: List[MappedField]

    # The ID can from single field or list of fields (composite key)
    id: List[MappedField]

    # Description of the element
    description: Optional[str] = None

    # Cache config
    cache_config: Optional[GraphElementCacheConfig] = None

    # ID of the from vertex, only applicable if the element is an edge
    from_id: Optional[List[MappedField]] = None

    # ID of the to vertex, only applicable if the element is an edge
    to_id: Optional[List[MappedField]] = None

    # Label of the from vertex, only applicable if the element is an edge
    from_label: Optional[str] = None

    # Label of the to vertex, only applicable if the element is an edge
    to_label: Optional[str] = None


@dataclass(frozen=True)
class PuppyGraphMappingConfig:
    """PuppyGraph Mapping config.

    This config defines how the graph is mapped from the source tables.
    """

    # Catalogs to fetch tables from
    catalogs: List[CatalogConfig] = field(default_factory=list)

    vertices: List[GraphElementConfig] = field(default_factory=list)

    edges: List[GraphElementConfig] = field(default_factory=list)

    # description of the graph
    description: Optional[str] = None


if __name__ == "__main__":
    pass
    # print(GraphElementConfig)
    # print(PuppyGraphConstruction



"""PuppyGraph Element Cache config."""

from dataclasses import dataclass
from enum import Enum
from typing import Optional


class TimeUnit(Enum):
    """Enum for time units in cache TTL."""

    YEAR = 0
    MONTH = 1
    DAY = 2
    HOUR = 3
    MINUTE = 4
    PARTITION = 5


@dataclass(frozen=True)
class TimeToLive:
    """Time-to-live configuration for cache."""

    amount: int
    unit: TimeUnit


@dataclass(frozen=True)
class RefreshStrategy:
    """Refresh strategy configuration for cache."""

    parallelism: int


@dataclass(frozen=True)
class GraphElementCacheConfig:
    """Cache configuration for elements."""

    partition_key: str
    partition_ttl: Optional[TimeToLive] = None
    refresh_strategy: Optional[RefreshStrategy] = None



"""Catalog configuration dataclasses and enums.

This module defines the access configurations for different types of data sources.
"""

from dataclasses import dataclass
from enum import Enum
from typing import Union

from puppygraph.data.mapping.database_params import DatabaseParams
from puppygraph.data.mapping.datalake_params import DatalakeParams


class CatalogType(Enum):
    """Defines the source type of a catalog."""

    UNKNOWN = 0
    HIVE = 1
    ICEBERG = 2
    HUDI = 3
    DELTALAKE = 4
    MYSQL = 5
    POSTGRESQL = 6
    ELASTICSEARCH = 7
    REDSHIFT = 8
    DUCKDB = 9
    BIGQUERY = 10
    SNOWFLAKE = 11
    TRINO = 12
    VERTICA = 13
    SINGLESTORE = 14


@dataclass(frozen=True)
class CatalogConfig:
    """Defines the source of a catalog for PuppyGraph to construct the graph from."""

    name: str
    type: CatalogType
    params: Union[DatalakeParams, DatabaseParams]



import pytest
from puppygraph.common.conversion_utils import convert_mapping_config_to_host_json


@pytest.mark.parametrize(
    "config_dict, expected_json",
    [
        (
            {
                "catalogs": [
                    {
                        "name": "my_catalog",
                        "type": "DELTALAKE",
                        "params": {
                            "metastore_param": {
                                "token": "my_token",
                                "host": "my_host",
                                "unity_catalog_name": "my_catalog_name",
                            },
                            "storage_param": {
                                "use_instance_profile": "false",
                                "region": "my_region",
                                "access_key": "my_access_key",
                                "secret_key": "my_secret_key",
                                "enable_ssl": "true",
                                "type": "S3",
                            },
                        },
                    }
                ],
                "vertices": [
                    {
                        "label": "Person",
                        "table_source": {
                            "catalog_name": "my_catalog",
                            "schema_name": "my_schema",
                            "table_name": "person_table",
                        },
                        "id": [{"name": "id", "type": "String", "from_field": "id"}],
                        "attributes": [
                            {
                                "name": "name",
                                "type": "String",
                                "from_field": "person_name",
                            },
                            {"name": "age", "type": "Int", "from_field": "person_age"},
                        ],
                    },
                    {
                        "label": "Location",
                        "table_source": {
                            "catalog_name": "my_catalog",
                            "schema_name": "my_schema",
                            "table_name": "location_table",
                        },
                        "id": [{"name": "id", "type": "String", "from_field": "id"}],
                        "attributes": [
                            {
                                "name": "name",
                                "type": "String",
                                "from_field": "location_name",
                            },
                            {
                                "name": "latitude",
                                "type": "Float",
                                "from_field": "location_lat",
                            },
                            {
                                "name": "longitude",
                                "type": "Float",
                                "from_field": "location_long",
                            },
                        ],
                    },
                ],
                "edges": [
                    {
                        "label": "LivesIn",
                        "from_label": "Person",
                        "to_label": "Location",
                        "table_source": {
                            "catalog_name": "my_catalog",
                            "schema_name": "my_schema",
                            "table_name": "lives_in_table",
                        },
                        "id": [{"name": "id", "type": "String", "from_field": "id"}],
                        "from_id": [
                            {
                                "name": "from_id",
                                "type": "String",
                                "from_field": "from_id",
                            }
                        ],
                        "to_id": [
                            {"name": "to_id", "type": "String", "from_field": "to_id"}
                        ],
                        "attributes": [
                            {
                                "name": "since",
                                "type": "Date",
                                "from_field": "since_date",
                            }
                        ],
                    },
                    {
                        "label": "Likes",
                        "from_label": "Person",
                        "to_label": "Person",
                        "table_source": {
                            "catalog_name": "my_catalog",
                            "schema_name": "my_schema",
                            "table_name": "likes_table",
                        },
                        "id": [{"name": "id", "type": "String", "from_field": "id"}],
                        "from_id": [
                            {
                                "name": "from_id",
                                "type": "String",
                                "from_field": "from_id",
                            }
                        ],
                        "to_id": [
                            {"name": "to_id", "type": "String", "from_field": "to_id"}
                        ],
                        "attributes": [],
                    },
                ],
            },
            {
                "catalogs": [
                    {
                        "name": "my_catalog",
                        "type": "deltalake",
                        "metastore": {
                            "type": "unity",
                            "token": "my_token",
                            "host": "my_host",
                            "databricksCatalogName": "my_catalog_name",
                        },
                        "storage": {
                            "useInstanceProfile": "false",
                            "region": "my_region",
                            "accessKey": "my_access_key",
                            "secretKey": "my_secret_key",
                            "enableSsl": "true",
                            "type": "S3",
                        },
                    },
                ],
                "graph": {
                    "vertices": [
                        {
                            "label": "Person",
                            "oneToOne": {
                                "tableSource": {
                                    "catalog": "my_catalog",
                                    "schema": "my_schema",
                                    "table": "person_table",
                                },
                                "id": {
                                    "fields": [
                                        {"field": "id", "type": "String", "alias": "id"}
                                    ]
                                },
                                "attributes": [
                                    {
                                        "alias": "name",
                                        "field": "person_name",
                                        "type": "String",
                                    },
                                    {
                                        "alias": "age",
                                        "field": "person_age",
                                        "type": "Int",
                                    },
                                ],
                            },
                        },
                        {
                            "label": "Location",
                            "oneToOne": {
                                "tableSource": {
                                    "catalog": "my_catalog",
                                    "schema": "my_schema",
                                    "table": "location_table",
                                },
                                "id": {
                                    "fields": [
                                        {"field": "id", "type": "String", "alias": "id"}
                                    ]
                                },
                                "attributes": [
                                    {
                                        "alias": "name",
                                        "field": "location_name",
                                        "type": "String",
                                    },
                                    {
                                        "alias": "latitude",
                                        "field": "location_lat",
                                        "type": "Float",
                                    },
                                    {
                                        "alias": "longitude",
                                        "field": "location_long",
                                        "type": "Float",
                                    },
                                ],
                            },
                        },
                    ],
                    "edges": [
                        {
                            "label": "LivesIn",
                            "fromVertex": "Person",
                            "toVertex": "Location",
                            "tableSource": {
                                "catalog": "my_catalog",
                                "schema": "my_schema",
                                "table": "lives_in_table",
                            },
                            "id": {
                                "fields": [
                                    {"field": "id", "type": "String", "alias": "id"}
                                ]
                            },
                            "fromId": {
                                "fields": [
                                    {
                                        "field": "from_id",
                                        "type": "String",
                                        "alias": "from_id",
                                    }
                                ]
                            },
                            "toId": {
                                "fields": [
                                    {
                                        "field": "to_id",
                                        "type": "String",
                                        "alias": "to_id",
                                    }
                                ]
                            },
                            "attributes": [
                                {
                                    "alias": "since",
                                    "field": "since_date",
                                    "type": "Date",
                                }
                            ],
                        },
                        {
                            "label": "Likes",
                            "fromVertex": "Person",
                            "toVertex": "Person",
                            "tableSource": {
                                "catalog": "my_catalog",
                                "schema": "my_schema",
                                "table": "likes_table",
                            },
                            "id": {
                                "fields": [
                                    {"field": "id", "type": "String", "alias": "id"}
                                ]
                            },
                            "fromId": {
                                "fields": [
                                    {
                                        "field": "from_id",
                                        "type": "String",
                                        "alias": "from_id",
                                    }
                                ]
                            },
                            "toId": {
                                "fields": [
                                    {
                                        "field": "to_id",
                                        "type": "String",
                                        "alias": "to_id",
                                    }
                                ]
                            },
                            "attributes": [],
                        },
                    ],
                },
            },
        )
    ],
)
def test_convert_schema_and_construction_to_host_json(config_dict, expected_json):
    """Test conversion of schema and construction to host json."""
    assert convert_mapping_config_to_host_json(config=config_dict) == expected_json



from puppygraph.client.client import PuppyGraphClient
from puppygraph.data.host.host_config import PuppyGraphHostConfig
from puppygraph.data.schema.graph_schema import PuppyGraphSchema

__all__ = ["PuppyGraphClient", "PuppyGraphHostConfig", "PuppyGraphSchema"]



"""Defines Datalake parameters for a catalog."""

from dataclasses import dataclass
from typing import Optional, Union


@dataclass(frozen=True)
class UnityMetastoreParam:
    """Unity metastore parameters."""

    token: Optional[str] = None
    host: Optional[str] = None
    unity_catalog_name: Optional[str] = None


@dataclass(frozen=True)
class S3StorageParam:
    """S3 storage parameters."""

    use_instance_profile: Optional[str] = None
    region: Optional[str] = None
    access_key: Optional[str] = None
    secret_key: Optional[str] = None
    iam_role_arn: Optional[str] = None
    enable_ssl: Optional[str] = None
    endpoint: Optional[str] = None
    enable_path_style_access: Optional[str] = None


MetastoreParam = Union[UnityMetastoreParam]
StorageParam = Union[S3StorageParam]


@dataclass(frozen=True)
class DatalakeParams:
    """Datalake parameters for a catalog."""

    metastore_param: Optional[MetastoreParam] = None
    storage_param: Optional[StorageParam] = None



"""Config for PuppyGraph server."""

from dataclasses import dataclass


@dataclass(frozen=True)
class PuppyGraphHostConfig:
    """PuppyGraph host configuration."""

    # IP address of the PuppyGraph host
    ip: str

    # HTTP port of the PuppyGraph server
    http_port: int = 8081

    # Cypher query port of the PuppyGraph server
    cypher_port: int = 7687
    
    # Maximum connection lifetime for the Cypher query driver
    cypher_max_connection_lifetime: int = 30

    # Gremlin query port of the PuppyGraph server
    gremlin_port: int = 8182

    # Username to authenticate with the PuppyGraph server
    username: str = "puppygraph"

    # Password to authenticate with the PuppyGraph server
    password: str = "puppygraph123"



"""PuppyGraph client module."""

import logging
import threading
import time
from typing import Any, Dict, List, Optional, Union

import requests
from gremlin_python.driver import client as GremlinClient
from gremlin_python.driver.protocol import GremlinServerError
from gremlin_python.driver.serializer import GraphBinarySerializersV1
from neo4j import Driver as CypherDriver
from neo4j import GraphDatabase, Query, Result
from neo4j.exceptions import AuthError, CypherSyntaxError, ServiceUnavailable
from neo4j.graph import Node, Relationship, Path
from puppygraph.common.conversion_utils import convert_mapping_config_to_host_json
from puppygraph.data.host.host_config import PuppyGraphHostConfig
from puppygraph.data.mapping.graph_mapping_config import PuppyGraphMappingConfig

logger = logging.getLogger(__name__)
logger.addHandler(
    logging.NullHandler()
)  # Prevent "No handlers could be found" warnings


class PuppyGraphClient:
    """PuppyGraph client."""

    def __init__(self, config: PuppyGraphHostConfig):
        """Initializes a PuppyGraph client."""
        self._config = config
        self._cypher_query_driver = GraphDatabase.driver(
            f"bolt://{config.ip}:{config.cypher_port}/",
            auth=(config.username, config.password),
            max_connection_lifetime=config.cypher_max_connection_lifetime,
        )
        self._gremlin_query_client = GremlinClient.Client(
            f"ws://{config.ip}:{config.gremlin_port}/gremlin",
            traversal_source="g",
            username=config.username,
            password=config.password,
            message_serializer=GraphBinarySerializersV1(),
        )

    def set_schema(
        self, mapping_config: Union[PuppyGraphMappingConfig, Dict]
    ) -> PuppyGraphMappingConfig:
        """Sets the graph mapping config in PuppyGraph server.

        Args:
            mapping_config: The graph mapping config to set, can be json or dataclass.

        Returns:
            The puppygraph graph mapping config that was set.
        """
        return _set_schema(host_config=self._config, mapping_config=mapping_config)

    def get_schema(self) -> str:
        """Returns the json schema of the PuppyGraph database."""
        return _get_schema(config=self._config)

    def cypher_query(
        self, query: str, params: Optional[Dict[str, Any]] = None, timeout_ms=30000
    ) -> List[Dict[str, Any]]:
        """Executes a Cypher query on the puppy graph.

        Args:
            query: The Cypher query to execute.
            params: The parameters to pass to the Cypher query.
            timeout_ms: The timeout in milliseconds for the query, defaults to 30000.

        Returns:
            The result of the Cypher query in a list of dictionaries.
        """
        return _run_cypher_query(
            cypher_driver=self._cypher_query_driver,
            query=query,
            params=params,
            timeout_ms=timeout_ms,
        )

    def gremlin_query(self, query: str, timeout_ms=30000) -> List[Dict[str, Any]]:
        """Executes a Gremlin query on the puppy graph.

        Args:
            query: The Gremlin query to execute.
            timeout_ms: The timeout in milliseconds for the query, defaults to 30000.

        Returns:
            The result of the Gremlin query in a list of dictionaries.
        """
        return _run_gremlin_query(
            gremlin_client=self._gremlin_query_client,
            query=query,
            timeout_ms=timeout_ms,
        )


def _set_schema(
    host_config: PuppyGraphHostConfig,
    mapping_config: Union[PuppyGraphMappingConfig, Dict],
) -> PuppyGraphMappingConfig:
    """Sets the graph schema in PuppyGraph server.

    Args:
        host_config: The host configuration of the PuppyGraph server.
        mapping_config: The graph mapping config that defines how the graph is constructed, can be json or dataclass.

    Returns:
        The puppygraph schema that was set.
    """

    schema_json = convert_mapping_config_to_host_json(config=mapping_config)
    logger.info(
        "Setting graph schema in PuppyGraph server...\n=============================\n%s\n=============================\n",
        schema_json,
    )

    response = requests.post(
        f"http://{host_config.ip}:{host_config.http_port}/schema",
        auth=(host_config.username, host_config.password),
        json=schema_json,
        timeout=60,
    )
    response.raise_for_status()
    logger.info("Successfully updated the schema in PuppyGraph!")

    # Check if PuppyGraph is ready to serve
    response = requests.get(
        f"http://{host_config.ip}:{host_config.http_port}/schemajson",
        auth=(host_config.username, host_config.password),
        timeout=60,
    )
    while response.status_code != 200:
        time.sleep(10)
    logger.info("PuppyGraph is ready to serve the new schema!")
    return mapping_config


def _get_schema(config: PuppyGraphHostConfig) -> str:
    """Returns the schema of the PuppyGraph database.

    Args:
        config: The PuppyGraph host configuration.

    Returns:
        The schema of the PuppyGraph database in string format.
    """
    response = requests.get(
        f"http://{config.ip}:{config.http_port}/schemajson",
        auth=(config.username, config.password),
        timeout=60,
    )
    response.raise_for_status()
    return response.text


class _QueryThread(threading.Thread):
    def __init__(self, target, *args, **kwargs):
        super().__init__()
        self._target = target
        self._args = args
        self._kwargs = kwargs
        self._result = None
        self._error = None

    def run(self):
        try:
            self._result = self._target(*self._args, **self._kwargs)
        except Exception as e:
            self._error = e

    def get_result(self):
        if self._error:
            raise self._error
        return self._result


def _run_with_threading_timeout(fn, timeout, *args, **kwargs):
    thread = _QueryThread(fn, *args, **kwargs)
    thread.start()
    thread.join(timeout / 1e3)  # Convert ms to seconds

    if thread.is_alive():
        raise TimeoutError(f"Operation timed out after {timeout} ms!")
    return thread.get_result()


def _cypher_query_fn(
    cypher_driver: CypherDriver,
    query: str,
    params: Optional[Dict[str, Any]],
    timeout_s: float,
):
    with cypher_driver.session() as session:
        neo4j_query = Query(text=query, timeout=timeout_s)
        try:
            res: Result = session.run(neo4j_query, params)
            json_data: List[Dict[str, Any]] = [_unpack_value(record) for record in res]
            return json_data
        except CypherSyntaxError as e:
            raise ValueError(f"`{query}` is not valid:\n{e}") from e
        except (AuthError, ServiceUnavailable, TimeoutError) as e:
            raise e

def _unpack_value(value):
    if isinstance(value, Node):
        return {
            "type": "Node",
            "id": value.element_id,
            "labels": list(value.labels),
            "properties": dict(value.items())
        }
    elif isinstance(value, Relationship):
        return {
            "type": "Relationship",
            "id": value.element_id,
            "type_name": value.type,
            "start_id": value.start_node.element_id,
            "end_id": value.end_node.element_id,
            "properties": dict(value.items())
        }
    elif isinstance(value, Path):
        return {
            "type": "Path",
            "nodes": [_unpack_value(n) for n in value.nodes],
            "relationships": [_unpack_value(r) for r in value.relationships]
        }
    elif isinstance(value, list):
        return [_unpack_value(v) for v in value]
    elif isinstance(value, dict):
        return {k: _unpack_value(v) for k, v in value.items()}
    else:
        return value  # fall back to raw value (int, str, etc.)


def _gremlin_query_fn(gremlin_client: GremlinClient.Client, query: str):
    try:
        result_set = gremlin_client.submit(query)
        results = result_set.all().result()
        return results
    except GremlinServerError as e:
        raise ValueError(f"Gremlin query error: {e}") from e
    except TimeoutError as e:
        raise TimeoutError(f"Timeout occurred: {e}") from e
    except Exception as e:
        raise e


def _run_cypher_query(
    cypher_driver: CypherDriver,
    query: str,
    params: Optional[Dict[str, Any]] = None,
    timeout_ms: int = 300000,
) -> List[Dict[str, Any]]:
    return _run_with_threading_timeout(
        _cypher_query_fn, timeout_ms, cypher_driver, query, params, timeout_ms / 1e3
    )


def _run_gremlin_query(
    gremlin_client: GremlinClient.Client,
    query: str,
    timeout_ms: int = 300000,
) -> List[Dict[str, Any]]:
    return _run_with_threading_timeout(
        _gremlin_query_fn, timeout_ms, gremlin_client, query
    )



"""Utility functions for working with dataclasses."""

from dataclasses import asdict
from enum import Enum


def dataclass_to_camel_dict(instance) -> dict:
    """Convert a dataclass instance to a camelCase dictionary, omitting fields with None values."""
    if not hasattr(instance, "__dataclass_fields__"):
        raise ValueError("The provided instance is not a dataclass.")

    def _snake_to_camel(snake_str: str) -> str:
        """Convert a snake_case string to camelCase."""
        components = snake_str.split("_")
        return components[0] + "".join(x.title() for x in components[1:])

    def _convert_value(value):
        """Convert Enum to its name, and keep other types as they are."""
        if isinstance(value, Enum):
            return value.name
        return value

    return {
        _snake_to_camel(k): _convert_value(v)
        for k, v in asdict(instance).items()
        if v is not None
    }



"""PuppyGraphAgent."""

import logging
from copy import deepcopy
from typing import Iterable, List, Optional

from langchain_core.language_models.chat_models import BaseChatModel
from langchain_core.messages import (
    AIMessage,
    BaseMessage,
    HumanMessage,
    ToolCall,
    ToolMessage,
)
from langchain_core.prompts.chat import ChatPromptTemplate
from langchain_core.pydantic_v1 import Field, create_model
from langchain_core.runnables import RunnableSequence
from langchain_core.tools import StructuredTool, Tool

from puppygraph.client.client import PuppyGraphClient

logger = logging.getLogger(__name__)
logger.addHandler(
    logging.NullHandler()
)  # Prevent "No handlers could be found" warnings


class PuppyGraphAgent:
    """PuppyGraphAgent is the agent that interacts with the PuppyGraph via natural language.

    It enables the user to interact with the graph via natural lanaguage queries.
    """

    def __init__(
        self,
        puppy_graph_client: PuppyGraphClient,
        llm: BaseChatModel,
        chat_prompt_template: ChatPromptTemplate,
        query_language: str = "cypher",
        additional_tools: Optional[List[Tool]] = None,
    ):
        """Initializes the PuppyGraphAgent.

        Args:
            puppy_graph_client: The PuppyGraph client
            llm: The language model
            chat_prompt_template: The chat prompt template
            query_language: The query language to use, either "cypher" or "gremlin" or "both"
            additional_tools: Additional tools for the agent to use
        """

        self._puppy_graph_client = puppy_graph_client

        # Set up tools
        self._cypher_query_tool = _get_cypher_query_tool(
            puppy_graph_client=puppy_graph_client
        )
        self._gremlin_query_tool = _get_gremlin_query_tool(
            puppy_graph_client=puppy_graph_client
        )
        if query_language == "cypher":
            self._tools = [self._cypher_query_tool]
        elif query_language == "gremlin":
            self._tools = [self._gremlin_query_tool]
        elif query_language == "both":
            self._tools = [self._cypher_query_tool, self._gremlin_query_tool]

        if additional_tools is not None:
            self._tools.extend(additional_tools)

        self._tool_dict = {tool.name: tool for tool in self._tools}

        # Set up llm chain
        self._chat_prompt_template = chat_prompt_template

        self._llm = llm.bind_tools(tools=self._tools)
        self._llm_chain: RunnableSequence = self._chat_prompt_template | self._llm

        self._llm_no_tool_output = llm.bind_tools(tools=[])
        self._llm_no_tool_output_chain: RunnableSequence = (
            self._chat_prompt_template | self._llm_no_tool_output
        )

        # Set up other global variables
        self._message_history = []

    def query(self, user_input: str, max_iters: int = 10) -> Iterable[BaseMessage]:
        """Query the graph using the given user input.

        Args:
            user_input: The user input
            max_iters: The maximum number of iterations to run

        Yields:
            BaseMessage, can be either AIMessage or ToolMessage
        """
        # We have to copy the message history to avoid side effects
        # if query() is called multiple times
        message_history = deepcopy(self._message_history)

        new_messages = [HumanMessage(content=user_input)]

        iters = 0

        wait_for_user_input = False
        while iters < max_iters and not wait_for_user_input:
            # If we are at the last iteration, we don't want to call the tool
            # This is because we want to show the user the final message
            tool_call_allowed = True
            if iters + 1 == max_iters:
                tool_call_allowed = False

            # Predict
            ai_message = self._llm_predict(
                message_history=message_history + new_messages,
                tool_call_allowed=tool_call_allowed,
            )

            # Add AI message to new messages
            new_messages.append(ai_message)
            logger.info("AI message: %s", ai_message.content)
            yield ai_message

            if ai_message.tool_calls:
                # Execute the actual tool
                tool_messages = self._execute_tool_calls(ai_message)

                # Add tool messages to new messages
                new_messages.extend(tool_messages)
                logger.info("Tool messages: %s", tool_messages)
                yield from tool_messages

            iters += 1

            # Check if we need to wait for user input
            if not ai_message.tool_calls:
                wait_for_user_input = True

        # Update the message history
        self._message_history += new_messages

    def reset_messages(self):
        """Reset the message history."""
        self._message_history = []

    def _llm_predict(
        self,
        message_history: List[BaseMessage],
        tool_call_allowed: bool,
    ) -> AIMessage:
        """Predict the AI message using llm.

        Args:
            message_history: The message history
            tool_call_allowed: Whether tool calls are allowed

        Returns:
            The predicted AI message
        """
        input_dict = {
            "message_history": message_history,
        }

        if not tool_call_allowed:
            return self._llm_no_tool_output_chain.invoke(input=input_dict)

        return self._llm_chain.invoke(input=input_dict)

    def _execute_tool_calls(self, ai_message: AIMessage) -> List[ToolMessage]:
        """Execute the tool calls in the AI message.

        Args:
            ai_message: The AI message which might contain tool calls

        Returns:
            The tool messages
        """
        tool_messages = []
        for tool_call in ai_message.tool_calls:
            tool_messages.append(self._execute_tool_call(tool_call))
        return tool_messages

    def _execute_tool_call(self, tool_call: ToolCall) -> ToolMessage:
        """Execute the given tool call.

        Args:
            tool_call: The tool call to execute

        Returns:
            The tool message
        """
        tool = self._tool_dict[tool_call["name"]]
        logger.info(
            "Calling tool: %s with args: %s", tool_call["name"], tool_call["args"]
        )
        try:
            tool_output = str(tool.invoke(input=tool_call["args"]))
        except Exception as e:
            tool_output = f"While executing tool, an error occurred : {e}"

        return ToolMessage(tool_output, tool_call_id=tool_call["id"])


def _get_cypher_query_tool(puppy_graph_client: PuppyGraphClient):
    """Get the Cypher query tool."""
    return StructuredTool.from_function(
        func=puppy_graph_client.cypher_query,
        name="query_graph_cypher",
        description="Query the graph database using Cypher.",
        args_schema=create_model(
            "", query=(str, Field(description="The Cypher query to run"))
        ),
    )


def _get_gremlin_query_tool(puppy_graph_client: PuppyGraphClient):
    """Get the Gremlin query tool."""
    return StructuredTool.from_function(
        func=puppy_graph_client.gremlin_query,
        name="query_graph_gremlin",
        description="Query the graph database using Gremlin.",
        args_schema=create_model(
            "", query=(str, Field(description="The Gremlin query to run"))
        ),
    )



"""PuppyGraph schema definition."""

from dataclasses import dataclass, field
from typing import List, Optional


@dataclass(frozen=True)
class AttributeSchema:
    """Attribute schema."""

    type: str
    name: str
    description: Optional[str] = None


@dataclass(frozen=True)
class VertexSchema:
    """Vertex schema."""

    label: str
    attributes: List[AttributeSchema] = field(default_factory=list)
    description: Optional[str] = None


@dataclass(frozen=True)
class EdgeSchema:
    """Edge schema."""

    label: str
    from_vertex_label: str
    to_vertex_label: str
    attributes: List[AttributeSchema] = field(default_factory=list)
    description: Optional[str] = None


@dataclass(frozen=True)
class PuppyGraphSchema:
    """The PuppyGraph schema."""

    vertex_schemas: List[VertexSchema] = field(default_factory=list)
    edge_schemas: List[EdgeSchema] = field(default_factory=list)
    description: Optional[str] = None
